# Collection of configurable development workflows
# Copyright (C) 2011-2019 The Roberto Development Team
#
# This file is part of Roberto.
#
# Roberto is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Roberto is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --
"""Utilities used by tasks in Roberto's workflow."""


import hashlib
import os
import re
from typing import List

from invoke import Context, Failure


def parse_git_describe(git_describe: str) -> dict:
    """Parse the output of ``git describe --tags``.

    Parameters
    ----------
    git_describe
        The output of git describe

    Returns
    -------
    version_info: dict
        A dictionary with version information.

    """
    # Split the input string into basic parts.
    git_describe = git_describe.strip()
    version_words = git_describe.split('-')
    tag = version_words[0]
    version_parts = tag.split('.')
    if len(version_parts) != 3:
        raise TagError('A version tag must at least contain two dots.', tag)

    # If the last version part has a non numeric suffix,
    # it needs to be taken out.
    match = re.fullmatch(r'(\d+)(.*)', version_parts[-1])
    if match is None:
        raise TagError('The patch version must start with a number.', tag)
    version_parts[-1] = match.group(1)
    suffix = match.group(2)

    # Define various meaningful pieces
    major, minor, patch = version_parts
    if not minor.isnumeric():
        raise TagError('The minor version must be numeric.', tag)
    if not major.isnumeric():
        raise TagError('The major version must be numeric.', tag)
    if len(version_words) > 1:
        suffix += '.post' + version_words[1]

    # Construct results
    version_info = {
        'describe': git_describe,
        'tag': tag,
        'tag_version': f'{major}.{minor}.{patch}{suffix}',
        'tag_soversion': f'{major}.{minor}',
        'tag_version_major': major,
        'tag_version_minor': minor,
        'tag_version_patch': patch,
        'tag_version_suffix': suffix,
        'tag_stable': len(suffix) == 0,
        'tag_test': re.fullmatch(r'b\d+', suffix) is not None,
        'tag_dev': re.fullmatch(r'a\d+', suffix) is not None,
    }
    version_info['tag_release'] = (
        version_info['tag_stable']
        or version_info['tag_test']
        or version_info['tag_dev'])
    if version_info['tag_stable']:
        version_info['deploy_label'] = "main"
        version_info['dev_classifier'] = 'Development Status :: 5 - Production/Stable'
    elif version_info['tag_test']:
        version_info['deploy_label'] = "test"
        version_info['dev_classifier'] = 'Development Status :: 4 - Beta'
    elif version_info['tag_dev']:
        version_info['deploy_label'] = "dev"
        version_info['dev_classifier'] = 'Development Status :: 3 - Alpha'
    else:
        # Non-release tag or no tag.
        version_info['deploy_label'] = None
        version_info['dev_classifier'] = 'Development Status :: 2 - Pre-Alpha'
    return version_info


class TagError(Exception):
    """An exception raised when a version tag is invalid."""

    def __init__(self, message: str, tag: str):
        """Initialize an TagError.

        Parameters
        ----------
        message
            Describe the error.
        tag
            The tag having the problem, which will be added to the error
            message.

        """
        message = f'Invalid tag: {tag} ({message})'
        super().__init__(message)


def sanitize_branch(ctx: Context, branch: str):
    """Attempt to fix the presence of a branch.

    The branch is checked with rev-parse. If not present, try to set it to
    origin/branch. If that does not work, try to fetch it.

    Parameters
    ----------
    ctx
        A invoke.Context instance.
    branch
        The branch to resurrect.

    """
    # Test if merge branch is present
    try:
        ctx.run(f"git rev-parse --verify {branch}")
        return
    except Failure:
        print(f"Merge branch \"{branch}\" not found.")

    # Try to create it without connection to origin
    try:
        ctx.run(f"git branch --track {branch} origin/{branch}")
        return
    except Failure:
        print(f"Local copy of remote merge branch \"{branch}\" not found.")

    # Last resort: fetch the merge branch
    ctx.run(f"git fetch origin {branch}:{branch}")


def check_env_var(name: str):
    """Check if an environment variable is set and non-empty.

    Parameters
    ----------
    name
        The environment variable to be checked.

    """
    if name not in os.environ:
        return f'The environment variable {name} is not set.'
    if os.environ[name] == "":
        return f'The environment variable {name} is empty.'
    return f'The environment variable {name} is not empty.'


def need_deployment(ctx: Context, prefix: str, binary: bool, deploy_labels: List[str]) -> bool:
    """Return True if deployment is needed.

    Parameters
    ----------
    ctx
        A invoke.Context instance.
    prefix
        The type of deployment, used for message printed to stdout.
    binary
        Whether or not a binary is being deployment. If binary, it may be of
        interest to deploy for different architectures, which is controlled by
        a config.deploy_binary, intead of config.deploy_source.
    deploy_labels
        A list of valid deploy labels for this release. If the current deploy
        label matches any of the ones in the list, deployment is needed.

    """
    if binary and not ctx.deploy_binary:
        print(f"{prefix} not requested in configuration. (binary)")
        return False
    if not binary and not ctx.deploy_noarch:
        print(f"{prefix} not requested in configuration. (noarch)")
        return False
    if ctx.git.deploy_label not in deploy_labels:
        print(f"{prefix} skipped, because of deploy label {ctx.git.deploy_label}.")
        return False
    return True


def write_sha256_sum(fn_asset: str) -> str:
    """Make a sha256 checksum file, print it and return the checksum filename.

    Parameters
    ----------
    fn_asset
        The file of which the hash will be computed.

    Returns
    -------
    fn_asset_hash : str
        The filename of the file containing the hash.

    """
    hasher = hashlib.sha256()
    with open(fn_asset, 'br') as f:
        for chunck in iter(lambda: f.read(4096), b""):
            hasher.update(chunck)
    fn_sha256 = fn_asset + '.sha256'
    with open(fn_sha256, 'w') as f:
        line = f'{hasher.hexdigest()}  {fn_asset}'
        f.write(line + '\n')
        print(line)
    return fn_sha256



# Collection of configurable development workflows
# Copyright (C) 2011-2019 The Roberto Development Team
#
# This file is part of Roberto.
#
# Roberto is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Roberto is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --



# Collection of configurable development workflows
# Copyright (C) 2011-2019 The Roberto Development Team
#
# This file is part of Roberto.
#
# Roberto is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Roberto is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --
"""Unit tests Roberto.utils."""

import os

import pytest

from invoke.config import DataProxy

from ..utils import (parse_git_describe, write_sha256_sum, TagError,
                     check_env_var, need_deployment)


def test_parse_git_describe():
    assert parse_git_describe("1.2.3") == {
        'describe': '1.2.3',
        'tag': '1.2.3',
        'tag_version': '1.2.3',
        'tag_soversion': '1.2',
        'tag_version_major': '1',
        'tag_version_minor': '2',
        'tag_version_patch': '3',
        'tag_version_suffix': '',
        'tag_stable': True,
        'tag_test': False,
        'tag_dev': False,
        'tag_release': True,
        'deploy_label': 'main',
        'dev_classifier': 'Development Status :: 5 - Production/Stable',
    }
    assert parse_git_describe("0.18.1b1") == {
        'describe': '0.18.1b1',
        'tag': '0.18.1b1',
        'tag_version': '0.18.1b1',
        'tag_soversion': '0.18',
        'tag_version_major': '0',
        'tag_version_minor': '18',
        'tag_version_patch': '1',
        'tag_version_suffix': 'b1',
        'tag_stable': False,
        'tag_test': True,
        'tag_dev': False,
        'tag_release': True,
        'deploy_label': 'test',
        'dev_classifier': 'Development Status :: 4 - Beta',
    }
    assert parse_git_describe("30.1.2a7") == {
        'describe': '30.1.2a7',
        'tag': '30.1.2a7',
        'tag_version': '30.1.2a7',
        'tag_soversion': '30.1',
        'tag_version_major': '30',
        'tag_version_minor': '1',
        'tag_version_patch': '2',
        'tag_version_suffix': 'a7',
        'tag_stable': False,
        'tag_test': False,
        'tag_dev': True,
        'tag_release': True,
        'deploy_label': 'dev',
        'dev_classifier': 'Development Status :: 3 - Alpha',
    }
    assert parse_git_describe("15.13.11a9-10") == {
        'describe': '15.13.11a9-10',
        'tag': '15.13.11a9',
        'tag_version': '15.13.11a9.post10',
        'tag_soversion': '15.13',
        'tag_version_major': '15',
        'tag_version_minor': '13',
        'tag_version_patch': '11',
        'tag_version_suffix': 'a9.post10',
        'tag_stable': False,
        'tag_test': False,
        'tag_dev': False,
        'tag_release': False,
        'deploy_label': None,
        'dev_classifier': 'Development Status :: 2 - Pre-Alpha',
    }
    assert parse_git_describe("1.0.30l") == {
        'describe': '1.0.30l',
        'tag': '1.0.30l',
        'tag_version': '1.0.30l',
        'tag_soversion': '1.0',
        'tag_version_major': '1',
        'tag_version_minor': '0',
        'tag_version_patch': '30',
        'tag_version_suffix': 'l',
        'tag_stable': False,
        'tag_test': False,
        'tag_dev': False,
        'tag_release': False,
        'deploy_label': None,
        'dev_classifier': 'Development Status :: 2 - Pre-Alpha',
    }
    assert parse_git_describe("5.0.0") == {
        'describe': '5.0.0',
        'tag': '5.0.0',
        'tag_version': '5.0.0',
        'tag_soversion': '5.0',
        'tag_version_major': '5',
        'tag_version_minor': '0',
        'tag_version_patch': '0',
        'tag_version_suffix': '',
        'tag_stable': True,
        'tag_test': False,
        'tag_dev': False,
        'tag_release': True,
        'deploy_label': 'main',
        'dev_classifier': 'Development Status :: 5 - Production/Stable',
    }
    assert parse_git_describe("2.7.0-3-foo") == {
        'describe': '2.7.0-3-foo',
        'tag': '2.7.0',
        'tag_version': '2.7.0.post3',
        'tag_soversion': '2.7',
        'tag_version_major': '2',
        'tag_version_minor': '7',
        'tag_version_patch': '0',
        'tag_version_suffix': '.post3',
        'tag_stable': False,
        'tag_test': False,
        'tag_dev': False,
        'tag_release': False,
        'deploy_label': None,
        'dev_classifier': 'Development Status :: 2 - Pre-Alpha',
    }
    assert parse_git_describe("0.9.0a2") == {
        'describe': '0.9.0a2',
        'tag': '0.9.0a2',
        'tag_version': '0.9.0a2',
        'tag_soversion': '0.9',
        'tag_version_major': '0',
        'tag_version_minor': '9',
        'tag_version_patch': '0',
        'tag_version_suffix': 'a2',
        'tag_stable': False,
        'tag_test': False,
        'tag_dev': True,
        'tag_release': True,
        'deploy_label': 'dev',
        'dev_classifier': 'Development Status :: 3 - Alpha',
    }
    assert parse_git_describe("0.0.0-666-notag") == {
        'describe': '0.0.0-666-notag',
        'tag': '0.0.0',
        'tag_version': '0.0.0.post666',
        'tag_soversion': '0.0',
        'tag_version_major': '0',
        'tag_version_minor': '0',
        'tag_version_patch': '0',
        'tag_version_suffix': '.post666',
        'tag_stable': False,
        'tag_test': False,
        'tag_dev': False,
        'tag_release': False,
        'deploy_label': None,
        'dev_classifier': 'Development Status :: 2 - Pre-Alpha'
    }
    with pytest.raises(TagError):
        parse_git_describe('1.2')
    with pytest.raises(TagError):
        parse_git_describe('0.0.0.1')
    with pytest.raises(TagError):
        parse_git_describe('0.0.foo')
    with pytest.raises(TagError):
        parse_git_describe('0.foo.0')
    with pytest.raises(TagError):
        parse_git_describe('foo.0.0')


def test_check_env_var():
    # This test assumes the following variable is not defined in the environment.
    name = "d2f400557595b05b39dc6153567b2493c75b132114571b2bdd5d375c88ea732c"
    assert name not in os.environ
    assert check_env_var(name) == f'The environment variable {name} is not set.'
    os.environ[name] = ''
    assert check_env_var(name) == f'The environment variable {name} is empty.'
    os.environ[name] = '1'
    assert check_env_var(name) == f'The environment variable {name} is not empty.'


def test_need_deployment():
    ctx = DataProxy.from_data({
        'deploy_binary': False,
        'deploy_noarch': False,
        'git': {'deploy_label': 'foo'}})
    assert not need_deployment(ctx, '', True, ['foo', 'bar'])
    assert not need_deployment(ctx, '', True, ['egg', 'foo'])
    assert not need_deployment(ctx, '', True, ['bar', 'egg'])
    assert not need_deployment(ctx, '', False, ['foo', 'bar'])
    assert not need_deployment(ctx, '', False, ['egg', 'foo'])
    assert not need_deployment(ctx, '', False, ['bar', 'egg'])
    ctx = DataProxy.from_data({
        'deploy_binary': True,
        'deploy_noarch': False,
        'git': {'deploy_label': 'foo'}})
    assert need_deployment(ctx, '', True, ['foo', 'bar'])
    assert need_deployment(ctx, '', True, ['egg', 'foo'])
    assert not need_deployment(ctx, '', True, ['bar', 'egg'])
    assert not need_deployment(ctx, '', False, ['foo', 'bar'])
    assert not need_deployment(ctx, '', False, ['egg', 'foo'])
    assert not need_deployment(ctx, '', False, ['bar', 'egg'])
    ctx = DataProxy.from_data({
        'deploy_binary': False,
        'deploy_noarch': True,
        'git': {'deploy_label': 'foo'}})
    assert not need_deployment(ctx, '', True, ['foo', 'bar'])
    assert not need_deployment(ctx, '', True, ['egg', 'foo'])
    assert not need_deployment(ctx, '', True, ['bar', 'egg'])
    assert need_deployment(ctx, '', False, ['foo', 'bar'])
    assert need_deployment(ctx, '', False, ['egg', 'foo'])
    assert not need_deployment(ctx, '', False, ['bar', 'egg'])
    ctx = DataProxy.from_data({
        'deploy_binary': True,
        'deploy_noarch': True,
        'git': {'deploy_label': 'foo'}})
    assert need_deployment(ctx, '', True, ['foo', 'bar'])
    assert need_deployment(ctx, '', True, ['egg', 'foo'])
    assert not need_deployment(ctx, '', True, ['bar', 'egg'])
    assert need_deployment(ctx, '', False, ['foo', 'bar'])
    assert need_deployment(ctx, '', False, ['egg', 'foo'])
    assert not need_deployment(ctx, '', False, ['bar', 'egg'])


def test_write_sha256_sum_1(tmpdir):
    fn_test = os.path.join(str(tmpdir), 'a.bin')
    with open(fn_test, "wb") as f:
        f.write(b"foobar\n")
    fn_hash = write_sha256_sum(fn_test)
    assert os.path.isfile(fn_hash)
    with open(fn_hash, 'r') as f:
        assert f.read() == ("aec070645fe53ee3b3763059376134f058cc337247c978add1"
                            f"78b6ccdfb0019f  {fn_test}\n")


def test_write_sha256_sum_1000(tmpdir):
    fn_test = os.path.join(str(tmpdir), 'a.bin')
    with open(fn_test, "wb") as f:
        f.write(b"eggspam\n"*1000)
    fn_hash = write_sha256_sum(fn_test)
    assert os.path.isfile(fn_hash)
    os.system(f"sha256sum {fn_test}")
    with open(fn_hash, 'r') as f:
        assert f.read() == ("e3cf7ad45677ef171d77ec47e2dea492ba32e36b9d9fbcd0c9"
                            f"0951421d78bcc9  {fn_test}\n")



# Collection of configurable development workflows
# Copyright (C) 2011-2019 The Roberto Development Team
#
# This file is part of Roberto.
#
# Roberto is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Roberto is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --
"""Tools contribute functionality to tasks."""


from glob import glob
import json
import os
import urllib.request

from invoke import UnexpectedExit

from .utils import (sanitize_branch, need_deployment,
                    check_env_var, write_sha256_sum)
from .testenvs import append_activate


TOOL_CLASSES = {}
TOOLS = {}


def initialize_tools(tools_config):
    """Build a dictionary with available tools."""
    for tool_name, tool_config in tools_config.items():
        # Extract generic stuff
        cls = tool_config.pop("cls")
        kwargs = dict(tool_config)
        supported_envs = kwargs.pop("supported_envs", ["conda", "venv"])
        requirements = kwargs.pop("requirements", [])
        # Initialize
        tool = TOOL_CLASSES[cls](tool_name, **kwargs)
        # Inject attributes, to keep constructors light. Ugly here, but sweet
        # in code below.
        tool.supported_envs = supported_envs
        tool.requirements = requirements
        # Register
        TOOLS[tool_name] = tool


def execute_tools(ctx, tool_cls):
    """Execute all tools of a given class for all packages."""
    for package in ctx.project.packages:
        for toolname in package.tools:
            tool = TOOLS[toolname]
            if isinstance(tool, tool_cls):
                # Skip tools which are not supported
                print(f"\033[0;94m   TOOL  {tool.name}\033[0;0m")
                if ctx.testenv.use not in tool.supported_envs:
                    print(
                        f"Tool {toolname} skipped due to incompatibility with "
                        f"{ctx.testenv.use}.")
                    continue
                fmtkwargs = {'config': ctx.config, 'package': package}
                tool.execute(ctx, package, fmtkwargs)


class Tool:
    """Tool base class."""

    def __init_subclass__(cls, **kwargs):
        """Register any new sub class."""
        super().__init_subclass__(**kwargs)
        TOOL_CLASSES[cls.__name__] = cls

    def execute(self, ctx, package, fmtkwargs):
        """Execute the tool for a given package.

        Parameters
        ----------
        ctx
            Invoke Context instance.
        package
            Package configuration.
        fmtkwargs
            A dictionary used for formatting arguments.

        """
        raise NotImplementedError

    @staticmethod
    def _run_commands(ctx, package, fmtkwargs, commands):
        """Run all commengs for a given package, performanc variable substitution."""
        with ctx.cd(package.path), ctx.prefix(ctx.testenv.activate):
            for command in commands:
                ctx.run(command.format(**fmtkwargs))


class WriteVersion(Tool):
    """Write version info to a file."""

    def __init__(self, name, template, destination):
        """Initialize a WriteVersion instance.

        Parameters
        ----------
        name
            The name of the tool
        template
            Template for the source code with the version info.
        destination
            The destination to write the file too.

        """
        self.name = name
        self.template = template
        self.destination = destination

    def execute(self, ctx, package, fmtkwargs):
        """Execute the tool for a given package."""
        fn_version = self.destination.format(**fmtkwargs)
        content = self.template.format(**fmtkwargs)
        with open(os.path.join(package.path, fn_version), 'w') as f:
            f.write(content)
        print("Version file written to:", fn_version)


class Lint(Tool):
    """Base class for linters."""

    def __init__(self, name, commands_master, commands_feature):
        """Initialize a Lint instance.

        Parameters
        ----------
        name
            The name of the tool
        commands_master
            Executed when working in the merge_branch, usually master.
        commands_feature
            Executed when working in a feature branch.

        """
        self.name = name
        self.commands_master = commands_master
        self.commands_feature = commands_feature

    def execute(self, ctx, package, fmtkwargs):
        """Execute the tool for a given package."""
        if ctx.absolute:
            on_merge_branch = True
        else:
            result = ctx.run(f'git diff {ctx.git.merge_branch}..HEAD --stat', hide='out')
            on_merge_branch = (result.stdout.strip() == "")
        if on_merge_branch:
            self._run_commands(ctx, package, fmtkwargs, self.commands_master)
        else:
            self._run_commands(ctx, package, fmtkwargs, self.commands_feature)


class LintStatic(Lint):
    """Stub for static linters."""


class LintDynamic(Lint):
    """Stub for dynamic linters."""


class BuildInPlace(Tool):
    """Build in-place and check and/or extend environment variables."""

    def __init__(self, name, check_vars, commands, extra_vars):
        """Initialize a BuiltInPlace instance.

        Parameters
        ----------
        name
            The name of the tool
        check_vars
            A list of variable names to print before building, just to
            help detecting issues.
        commands
            Build commands.
        extra_vars
            Additions to environment variables needed after this build.

        """
        self.name = name
        self.check_vars = check_vars
        self.commands = commands
        self.extra_vars = extra_vars

    def _check_vars(self):
        """Print out some variables who might reveal issues."""
        print('Existing variables that could affect the in-place build:')
        for varname in self.check_vars:
            if varname in os.environ:
                print(f'{varname}={os.environ[varname]}')

    def _update_extra_vars(self, ctx, fmtkwargs):
        """Update environment variables after running build commands."""
        for name, value in self.extra_vars.items():
            separator = ":" if "PATH" in name else " "
            value = value.format(**fmtkwargs)
            append_activate(ctx, f'export {name}="${{{name}:+${{{name}}}{separator}}}{value}"')

    def execute(self, ctx, package, fmtkwargs):
        """Execute the tool for a given package."""
        self._check_vars()
        self._run_commands(ctx, package, fmtkwargs, self.commands)
        self._update_extra_vars(ctx, fmtkwargs)


class RunCommands(Tool):
    """Base class for all tools that just execute some commands."""

    def __init__(self, name, commands):
        """Initialize a BuiltInPlace instance.

        Parameters
        ----------
        name
            The name of the tool
        commands
            Commands to run.

        """
        self.name = name
        self.commands = commands

    def execute(self, ctx, package, fmtkwargs):
        """Execute the tool for a given package."""
        self._run_commands(ctx, package, fmtkwargs, self.commands)


class TestInPlace(RunCommands):
    """Stub for in-place testing tools."""


class UploadCoverage(RunCommands):
    """Stub for coverage upload."""


class BuildDocs(RunCommands):
    """Stub for documentation builds."""


class UploadDocs(Tool):
    """Base class for all docu upload tools."""


class UploadDocsGit(UploadDocs):
    """Squash-push documentation to a git branch."""

    def __init__(self, name, docroot, docbranch, docremote, deploy_labels):
        """Initialize an UploadDocsGit instance.

        Parameters
        ----------
        name
            The name of the tool
        docroot
            The subdirectory where the documentation can be found.
        docbranch
            The branch to which the documentation must be pushed.
        docremote
            The remote repository for the documentation.
        deploy_labels
            The kind of releases for which documentation must be uploaded.

        """
        self.name = name
        self.docroot = docroot
        self.docbranch = docbranch
        self.docremote = docremote
        self.deploy_labels = deploy_labels

    def execute(self, ctx, package, fmtkwargs):
        """Execute the tool for a given package."""
        # Try to get a git username and author argument for the doc commit.
        if 'GITHUB_TOKEN' in os.environ:
            # First get user info from the owner of the token
            req = urllib.request.Request('https://api.github.com/user')
            req.add_header('Authorization', f'token {os.environ["GITHUB_TOKEN"]}')
            try:
                with urllib.request.urlopen(req) as f:
                    user_info = json.loads(f.read().decode('utf-8'))['login']
            except urllib.error.HTTPError:
                # The exception may be raised when the token has no permission to
                # access user information.
                user_info = {}
            author_name = user_info.get("name", user_info.get("login", "Roberto"))
            fallback_email = user_info.get("login", "roberto") + "@users.noreply.github.com"
            author_email = user_info.get("email", fallback_email)
            os.environ["GIT_AUTHOR_NAME"] = author_name
            os.environ["GIT_AUTHOR_EMAIL"] = author_email
            os.environ["GIT_COMMITTER_NAME"] = author_name
            os.environ["GIT_COMMITTER_EMAIL"] = author_email

        # Check if deployment is needed with deploy_label.
        prefix = f'{self.name} of {package.dist_name}'
        if not need_deployment(ctx, prefix, False, self.deploy_labels):
            return

        with ctx.cd(package.path):
            # Switch to a docu branch and remove everything that was present in
            # the previous commit. It is assumed that the doc branch is an
            # orphan branch made previously.
            sanitize_branch(ctx, self.docbranch)
            ctx.run(f"git checkout {self.docbranch}")
            ctx.run("git ls-tree HEAD -r --name-only | xargs rm")
            # Copy the documentation to the repo root.
            docroot = self.docroot.format(**fmtkwargs)
            ctx.run(f"GLOBIGNORE='.:..'; cp -rv {docroot}/* .")
            # Add all files
            for root, _dirs, filenames in os.walk(docroot):
                for filename in filenames:
                    fullfn = os.path.join(root, filename)[len(docroot)+1:]
                    ctx.run(f"git add {fullfn}")
            # Commit, push and go back to the original branch
            ctx.run("git commit -a -m 'Automatic documentation update' --amend ")
            ctx.run(f"git checkout {ctx.git.branch}")
            # Push
            if 'GITHUB_TOKEN' in os.environ:
                # Get the remote url
                giturl = ctx.run("git config --get remote.origin.url").stdout.strip()
                # Derive the slug, works for both ssh and https
                slug = '/'.join(giturl.replace(':', '/').split('/')[-2:])
                # Push with github token magic. Taken from
                # https://gist.github.com/willprice/e07efd73fb7f13f917ea
                ctx.run(
                    "git remote add origin-pages "
                    f"https://{os.environ['GITHUB_TOKEN']}@github.com/{slug}",
                    hide=True, echo=False, warn=True
                )
                ctx.run(f"git push --quiet -f origin-pages {self.docbranch}:{self.docbranch}")
            else:
                # Fallback for local doc updates.
                ctx.run(f"git push -f {self.docremote} {self.docbranch}:{self.docbranch}")


class BuildPackage(RunCommands):
    """Stub for building a package."""


class Deploy(Tool):
    """Upload software packages to distribution servers."""

    def __init__(self, name, deploy_vars, include_sha256, noarch_asset_patterns,
                 binary_asset_patterns, deploy_labels, commands):
        """Initialize an UploadDocsGit instance.

        Parameters
        ----------
        name
            The name of the tool
        deploy_vars
            Essential environment variables for deployment. Their presence is
            checked.
        include_sha256
            If True, sha256 checksums are generated for the assets
        noarch_asset_patterns
            Glob pattern for architecture-independent assets.
        binary_asset_patterns
            Glob pattern for architecture-dependent assets.
        deploy_labels
            The kind of releases for which documentation must be uploaded.
        commands
            Bash command to perform the deployment.

        """
        self.name = name
        self.deploy_vars = deploy_vars
        self.include_sha256 = include_sha256
        self.noarch_asset_patterns = noarch_asset_patterns
        self.binary_asset_patterns = binary_asset_patterns
        self.deploy_labels = deploy_labels
        self.commands = commands

    def execute(self, ctx, package, fmtkwargs):
        """Execute the tool for a given package."""
        # Check if and how deployment vars are set.
        for deploy_var in self.deploy_vars:
            print(check_env_var(deploy_var))

        for binary, asset_patterns in [(True, self.binary_asset_patterns),
                                       (False, self.noarch_asset_patterns)]:
            # Fill in config variables in asset_patterns
            asset_patterns = [pattern.format(**fmtkwargs) for pattern in asset_patterns]
            descr = f'{self.name} of {package.dist_name} (binary={binary})'
            print(f"Preparing for {descr}")
            # Collect assets, skipping hash files previously generated.
            assets = []
            for pattern in asset_patterns:
                print("  Searching for", pattern)
                assets.extend([filename for filename in glob(pattern)
                               if not filename.endswith("sha256")])
            if not assets:
                print("No assets found")
                continue
            # Make sha256 checksums
            asset_hashes = [write_sha256_sum(asset) for asset in assets]
            if self.include_sha256:
                assets.extend(asset_hashes)
            # Print final assets
            print(f"Assets for upload: {assets}")
            # Set extra formatting variables.
            fmtkwargs['assets'] = ' '.join(assets)
            # Check if deployment is needed before running commands. This check
            # is maximally postponed to increase the coverage of the code above.
            if need_deployment(ctx, descr, binary, self.deploy_labels):
                try:
                    self._run_commands(ctx, package, fmtkwargs, self.commands)
                except UnexpectedExit:
                    print("\033[0;91m Deployment failed:\033[0;0m", descr)



# Collection of configurable development workflows
# Copyright (C) 2011-2019 The Roberto Development Team
#
# This file is part of Roberto.
#
# Roberto is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Roberto is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --
"""Installation of requirements."""


from glob import glob
import hashlib
import os
import tempfile
import time
from typing import Set

from invoke import Context
import yaml

from .tools import TOOLS


def compute_req_hash(req_items: Set[str], req_fns: Set[str]) -> str:
    """Compute a hash from all parameters that affect installed packages.

    Parameters
    ----------
    req_items
        A set of requirement items.
    req_fns
        A set of requirement files.

    Returns
    -------
    hashdigest : str
        The hex digest of the sha256 hash of all development requirements.

    """
    hasher = hashlib.sha256()
    for req_item in sorted(req_items):
        hasher.update(req_item.encode('utf-8'))
    for req_fn in sorted(req_fns):
        hasher.update(req_fn.encode("utf-8"))
        if os.path.isfile(req_fn):
            with open(req_fn, 'br') as f:
                hasher.update(f.read())
    return hasher.hexdigest()


def check_install_requirements(fn_skip: str, req_hash: str) -> bool:
    """Check if reinstallation of requirements is desired.

    Parameters
    ----------
    fn_skip
        File where the requirements hash is stored.
    req_has
        The (new) requirement hash.

    Returns
    -------
    install
        True if reinstallation is desired.

    """
    # The install and update will be skipped if it was done already once,
    # less than 24 hours ago and the req_hash has not changed.
    if os.path.isfile(fn_skip):
        if (time.time() - os.path.getmtime(fn_skip)) < 24*3600:
            with open(fn_skip) as f:
                if f.read().strip() == req_hash:
                    print("Skipping install+update of requirements.")
                    print(f"To force install+update: rm {fn_skip}")
                    return False
    print("Starting install+update of requirements.")
    print(f"To skip install+update: echo {req_hash} > {fn_skip}")
    return True


# pylint: disable=too-many-branches,too-many-statements
def install_requirements_conda(ctx: Context):
    """Install all requirements, including tools used by Roberto."""
    # Collect all parameters determining the install commands (to good
    # approximation) and turn them into a hash.
    # Some conda requirements are included by default because they must be present:
    # - conda: to make sure it is always up to date.
    # - conda-build: to have conda-render for getting requirements from recipes.
    conda_reqs = set(["conda"])
    pip_reqs = set([])
    recipe_dirs = []
    # Add project as a tool because it also contains requirements.
    tools = [ctx.project]
    for package in ctx.project.packages:
        for toolname in package.tools:
            tools.append(ctx.tools[toolname])
        recipe_dir = os.path.join(package.path, "tools", "conda.recipe")
        if os.path.isdir(recipe_dir):
            recipe_dirs.append(recipe_dir)
        else:
            print(f"Skipping recipe {recipe_dir}. (directory does not exist)")
    for tool in tools:
        for conda_req, pip_req in tool.get("requirements", []):
            if conda_req is None:
                pip_reqs.add(pip_req)
                conda_reqs.add("pip")
            else:
                conda_reqs.add(conda_req)
    req_hash = compute_req_hash(
        set("conda:" + conda_req for conda_req in conda_reqs) |
        set("pip:" + pip_req for pip_req in pip_reqs),
        sum([glob(os.path.join(recipe_dir, "*")) for recipe_dir in recipe_dirs], [])
    )

    fn_skip = os.path.join(ctx.testenv.path, ".skip_install")
    if check_install_requirements(fn_skip, req_hash):
        with ctx.prefix(ctx.conda.activate_base):
            # Update conda packages in the base env. Conda packages in the dev env
            # tend to be ignored.
            conda_reqs_base_str = " ".join(
                f"'{conda_req}'" for conda_req
                in conda_reqs if conda_req.startswith('conda')
            )
            ctx.run(f"conda install --update-deps -y {conda_reqs_base_str}")

        with ctx.prefix(ctx.testenv.activate):
            # Update packages already installed
            ctx.run("conda update --all -y")

            # Update and install other requirements for Roberto, in the dev env.
            conda_reqs_dev_str = " ".join(
                f"'{conda_req}'" for conda_req in conda_reqs
                if not conda_req.startswith('conda')
            )
            ctx.run(f"conda install --update-deps -y {conda_reqs_dev_str}")

            print("Rendering conda package, extracting requirements, which will be installed.")

            # Install dependencies from recipes, excluding own packages.
            own_conda_reqs = [package.dist_name for package in ctx.project.packages]
            for recipe_dir in recipe_dirs:
                # Send the output of conda render to a temporary directory.
                with tempfile.TemporaryDirectory() as tmpdir:
                    rendered_path = os.path.join(tmpdir, "rendered.yml")
                    ctx.run(
                        f"conda render -f {rendered_path} {recipe_dir} "
                        f"--variants {ctx.conda.variants}"
                    )
                    with open(rendered_path) as f:
                        rendered = yaml.safe_load(f)
                # Build a (simplified) list of requirements and install.
                dep_conda_reqs = set([])
                req_sources = [
                    ("requirements", 'build'),
                    ("requirements", 'host'),
                    ("requirements", 'run'),
                    ("test", 'requires'),
                ]
                for req_section, req_type in req_sources:
                    for recipe_req in rendered.get(req_section, {}).get(req_type, []):
                        words = recipe_req.split()
                        if words[0] not in own_conda_reqs:
                            dep_conda_req = words[0]
                            if len(words) > 1 and any(char in words[1] for char in "<>!="):
                                dep_conda_req += " " + words[1]
                            dep_conda_reqs.add(dep_conda_req)
                conda_reqs_render_str = " ".join(
                    f"'{conda_req}'" for conda_req in dep_conda_reqs
                )
                ctx.run(f"conda install --update-deps -y {conda_reqs_render_str}")

            # Update and install requirements for Roberto from pip, if any.
            if pip_reqs:
                pip_reqs_str = " ".join(f"'{pip_req}'" for pip_req in pip_reqs)
                ctx.run(f"pip install --upgrade {pip_reqs_str}")

        # Update the timestamp on the skip file.
        with open(fn_skip, 'w') as f:
            f.write(req_hash + '\n')


def convert_requires(fn_requires, fn_requirements):
    """Convert requires.txt from an egg-info to a requirements.txt file."""
    requires = []
    suffix = ""
    with open(fn_requires) as f:
        for line in f:
            line = line.strip()
            if len(line) == 0:
                continue
            if line.startswith("[:"):
                suffix = "; " + line[2:-1]
            else:
                requires.append(line + suffix)
    print("Extracted requirements:")
    with open(fn_requirements, "w") as f:
        for require in requires:
            print(require)
            f.write(require + "\n")


def install_requirements_pip(ctx: Context):
    """Install requirements in the virtual environment."""
    # Collect all parameters determining installation of requirements
    pip_reqs = set([])
    req_fns = set([])
    for _conda_req, pip_req in ctx.project.requirements:
        if pip_req is not None:
            pip_reqs.add(pip_req)
    for package in ctx.project.packages:
        for toolname in package.tools:
            tool = TOOLS[toolname]
            for _conda_req, pip_req in tool.requirements:
                if pip_req is not None:
                    pip_reqs.add(pip_req)
        req_fns.add(os.path.join(package.path, "setup.py"))
    req_hash = compute_req_hash(pip_reqs, req_fns)

    fn_skip = os.path.join(ctx.testenv.path, ".skip_install")
    if check_install_requirements(fn_skip, req_hash):
        with ctx.prefix(ctx.testenv.activate):
            if len(pip_reqs) > 0:
                # Upgrade pip
                ctx.run("pip install -U pip")
                # Install pip packages for the tools
                pip_reqs_str = " ".join(f"'{pip_req}'" for pip_req in pip_reqs)
                ctx.run(f"pip install -U {pip_reqs_str}")
            # Install dependencies for the project.
            for package in ctx.project.packages:
                with ctx.cd(package.path):
                    ctx.run("python setup.py egg_info")
                    fn_requires = os.path.join(
                        package.dist_name.replace("-", "_") + ".egg-info",
                        "requires.txt")
                    with tempfile.TemporaryDirectory() as tmpdir:
                        fn_requirements = os.path.join(tmpdir, "requirements.txt")
                        convert_requires(fn_requires, fn_requirements)
                        ctx.run("pip install -U -r " + fn_requirements)
        # Update the timestamp on the skip file.
        with open(fn_skip, 'w') as f:
            f.write(req_hash + '\n')



# Collection of configurable development workflows
# Copyright (C) 2011-2019 The Roberto Development Team
#
# This file is part of Roberto.
#
# Roberto is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Roberto is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --
"""Tools to set up virtual environments.

All classes implemented here are stateless. They do not have attributes and
store state information in the configuration object instead. They are
essentially groups of methods with compatible API, yet with implementation
details that may differ.
"""

import json
import os
import platform
import stat
import subprocess
import urllib.request

from invoke import Failure


def append_activate(ctx, line):
    """Append a line to the activate script and show it."""
    mode = "w" if ctx.testenv.activate == "true" else "a"
    with open(ctx.testenv.fn_activate, mode) as f:
        f.write(line + "\n")
    print(f"\033[0;96m   ENV:  {line}\033[0;0m")
    ctx.testenv.activate = "source " + os.path.abspath(ctx.testenv.fn_activate)


def install_macosx_sdk(ctx):
    """Install MacOSX SDK if on OSX if needed."""
    if platform.system() == 'Darwin' and ctx.macosx.install_sdk:
        optdir = os.path.join(ctx.testenv.base_path, 'opt')
        if not os.path.isdir(optdir):
            os.makedirs(optdir)
        sdk = f'MacOSX{ctx.macosx.release}.sdk'
        sdk_root = os.path.join(optdir, sdk)
        if not os.path.isdir(sdk_root):
            sdk_tar = f'{sdk}.tar.xz'
            sdk_dwnl = os.path.join(ctx.download_dir, sdk_tar)
            sdk_url = f'{ctx.maxosx.sdk_release}/{sdk_tar}'
            print(f"Downloading {sdk_url}")
            urllib.request.urlretrieve(sdk_url, sdk_dwnl)
            with ctx.cd(optdir):
                ctx.run(f'tar -xJf {sdk_dwnl}')
        append_activate(ctx, "export MACOSX_DEPLOYMENT_TARGET=" + ctx.macosx.release)
        append_activate(ctx, f'export SDKROOT="{sdk_root}"')
        print(f'MaxOSX sdk in: {sdk_root}')
        ctx.run(f'ls -alh {sdk_root}')
        ctx.macosx.sdk_root = sdk_root


def init_testenv(cfg):
    """Initialize a test environment defined in the configuration."""
    # Only initialize once
    if cfg.testenv.name is None:
        function = {
            "conda": init_testenv_conda,
            "venv": init_testenv_venv,
            "none": init_testenv_none,
        }[cfg.testenv.use]
        function(cfg)


def init_testenv_conda(cfg):
    """Initialize a testenv using Conda."""
    cfg.testenv = {}
    cfg.testenv.base_path = os.path.expandvars(os.path.expanduser(cfg.conda.base_path))
    cfg.testenv.name = cfg.project.name + '-dev'
    if cfg.conda.pinning:
        cfg.testenv.name += '-' + '-'.join(cfg.conda.pinning.split())
    cfg.testenv.path = os.path.join(cfg.testenv.base_path, 'envs', cfg.testenv.name)
    cfg.testenv.fn_activate = f"activate-conda-{cfg.testenv.name}.sh"
    cfg.testenv.activate = "true"
    cfg.testenv.setup = setup_conda
    cfg.testenv.nuke = nuke_conda


def install_conda(ctx):
    """Install miniconda if not present yet."""
    # Install miniconda if needed.
    if not os.path.isdir(os.path.join(ctx.testenv.base_path, 'bin')):
        dwnlconda = os.path.join(ctx.download_dir, 'miniconda.sh')
        if os.path.isfile(dwnlconda):
            print(f"Conda installer already present: {dwnlconda}")
        else:
            print(f"Downloading latest conda to {dwnlconda}.")
            if platform.system() == 'Darwin':
                urllib.request.urlretrieve(ctx.conda.osx_url, dwnlconda)
            elif platform.system() == 'Linux':
                urllib.request.urlretrieve(ctx.conda.linux_url, dwnlconda)
            else:
                raise Failure(f"Operating system {platform.system()} not supported.")

        # Fix permissions of the conda installer.
        os.chmod(dwnlconda, os.stat(dwnlconda).st_mode | stat.S_IXUSR)

        # Install
        print(f"Installing conda in {ctx.testenv.base_path}.")
        ctx.run(f"{dwnlconda} -b -p {ctx.testenv.base_path}")

    ctx.conda.activate_base = f"source '{ctx.testenv.base_path}/bin/activate'"

    install_macosx_sdk(ctx)


def setup_conda(ctx):
    """Set up a conda testing environment."""
    # Install stuff that may not be present.
    install_conda(ctx)

    # CONDA_BLD_PATH should not be overwritten, to allow for customization.
    if 'CONDA_BLD_PATH' in os.environ:
        ctx.conda.build_path = os.environ['CONDA_BLD_PATH']
    else:
        ctx.conda.build_path = os.path.join(ctx.testenv.path, 'conda-bld')

    # Check the sanity of the pinning configuration
    for char in "=<>!*":
        if char in ctx.conda.pinning:
            raise Failure(f"Character '{char}' should not be used in pinning.")
    pinned_words = ctx.conda.pinning.split()
    if len(pinned_words) % 2 != 0:
        raise Failure("Pinning config should be an even number of words, alternating "
                      "package names and versions, without wildcards.")

    # pinning as requirements for building the initial test environment.
    pinned_reqs = [
        f"{name}={version}" for name, version
        in zip(pinned_words[::2], pinned_words[1::2])]

    # Create the variants argument for render and build
    pinned_words = ctx.conda.pinning.split()
    ctx.conda.variants = '"{' + ','.join(
        f"{name}: '{version}'" for name, version
        in zip(pinned_words[::2], pinned_words[1::2])) + '}"'

    append_activate(ctx, '[[ -n \"${CONDA_PREFIX_1}\" ]] && conda deactivate &> /dev/null')
    append_activate(ctx, '[[ -n \"${CONDA_PREFIX}\" ]] && conda deactivate &> /dev/null')
    append_activate(ctx, ctx.conda.activate_base)

    with ctx.prefix(ctx.testenv.activate):
        # Check if the right environment exists, and make if needed.
        result = ctx.run("conda env list --json")
        print(f"Required conda env: {ctx.testenv.path}")
        if ctx.testenv.path not in json.loads(result.stdout)["envs"]:
            ctx.run(f"conda create -n {ctx.testenv.name} {' '.join(pinned_reqs)} -y")
            with open(os.path.join(ctx.testenv.path, "conda-meta", "pinning"), "w") as f:
                for pin in pinned_reqs:
                    f.write(pin + "\n")

    append_activate(ctx, f'conda activate {ctx.testenv.name}')
    append_activate(ctx, f'export CONDA_BLD_PATH="{ctx.conda.build_path}"')
    append_activate(ctx, f'export PROJECT_VERSION="{ctx.git.tag_version}"')

    with ctx.prefix(ctx.testenv.activate):
        # Reset the channels. Removing previous may fail if there were none. That's ok.
        ctx.run("conda config --env --remove-key channels", warn=True, hide='err')
        for channel in ctx.conda.channels:
            ctx.run(f"conda config --env --add channels {channel}")
        ctx.run("conda config --env --set channel_priority strict")


def nuke_conda(ctx):
    """Erase the conda environment."""
    # Go back to the base env before nuking the development env.
    setup_conda(ctx)
    with ctx.prefix(ctx.conda.activate_base):
        ctx.run(f"conda uninstall -n {ctx.testenv.name} --all -y")
    ctx.run("git clean -fdX")


def init_testenv_venv(cfg):
    """Initialize a test environment using Python's built-in venv."""
    cfg.testenv = {}
    cfg.testenv.base_path = os.path.expandvars(os.path.expanduser(cfg.venv.base_path))
    try:
        pyver = subprocess.run(
            ['python', '-c', 'import sys; print("{}.{}".format(*sys.version_info[:2]))'],
            stdout=subprocess.PIPE, stderr=subprocess.DEVNULL,
            check=True).stdout.decode('utf-8').strip()
    except subprocess.CalledProcessError:
        # May fail, e.g. when python binary is absent.
        pyver = 'X.Y'
    cfg.testenv.name = cfg.project.name + '-dev-python-' + pyver
    cfg.testenv.path = os.path.join(cfg.testenv.base_path, cfg.testenv.name)
    cfg.testenv.fn_activate = f"activate-venv-{cfg.testenv.name}.sh"
    cfg.testenv.activate = "true"
    cfg.testenv.setup = setup_venv
    cfg.testenv.nuke = nuke_venv


def setup_venv(ctx):
    """Make sure there is a virtual environment and activate it."""
    # Check if the required environment already exists
    if not os.path.isdir(ctx.testenv.path):
        # Create a new environment
        ctx.run(f"{ctx.venv.python_bin} -m venv {ctx.testenv.path}")
    else:
        print("Virtual environment already exists:")
        print(ctx.testenv.path)
    append_activate(ctx, '[[ -n "${VIRTUAL_ENV}" ]] && deactivate &> /dev/null')
    append_activate(ctx, f"source {ctx.testenv.path}/bin/activate")

    install_macosx_sdk(ctx)


def nuke_venv(ctx):
    """Erase the virtual environment."""
    # Not (yet) removing it, just showing how.
    ctx.run(f"echo rm -rv {ctx.testenv.path}")
    ctx.run("git clean -fdX")


def init_testenv_none(cfg):
    """Initialize a testenv without really doing anything."""
    cfg.testenv = {}
    cfg.testenv.base_path = None
    cfg.testenv.name = "noenv"
    cfg.testenv.path = None
    cfg.testenv.fn_activate = "activate.sh"
    cfg.testenv.activate = "true"
    cfg.testenv.setup = setup_none
    cfg.testenv.nuke = nuke_none


def setup_none(ctx):  # pylint: disable=unused-argument
    """Do nothing. Stub for API consistency."""


def nuke_none(ctx):  # pylint: disable=unused-argument
    """Do nothing. Stub for API consistency."""



# Collection of configurable development workflows
# Copyright (C) 2011-2019 The Roberto Development Team
#
# This file is part of Roberto.
#
# Roberto is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Roberto is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --
"""Tasks in Roberto's workflow.

Due to the decorators, no documentation is generated for this module. Use
``rob --help`` to get an overview of Roberto's task.
"""


from functools import wraps

from invoke import task

from .requirements import install_requirements_conda, install_requirements_pip
from .utils import sanitize_branch
from .tools import (execute_tools, WriteVersion, LintStatic, BuildInPlace,
                    TestInPlace, UploadCoverage, LintDynamic, BuildDocs,
                    UploadDocs, BuildPackage, Deploy)


TASK_HEADER = """\
\033[0;94m +-{line}-+\033[0;0m
\033[0;94m | TASK  {task} |\033[0;0m
\033[0;94m +-{line}-+\033[0;0m\
"""


def with_stdout_header(f):
    """Print a header for a task."""
    @wraps(f)
    def wrapper(ctx, *args, **kwargs):
        name = f.__name__
        print(TASK_HEADER.format(task=name, line="-"*(len(name)+6)))
        return f(ctx, *args, **kwargs)
    return wrapper


@task()
@with_stdout_header
def sanitize_git(ctx):
    """Fetch the git branch to be merged into, if it is absent."""
    sanitize_branch(ctx, ctx.git.merge_branch)
    print("Current branch:", ctx.git.branch)
    print("Merge branch:", ctx.git.merge_branch)


@task()
@with_stdout_header
def write_version(ctx):
    """Derive the version files from ``git describe --tags``."""
    execute_tools(ctx, WriteVersion)


@task()
@with_stdout_header
def setup_testenv(ctx):
    """Set up a testing environment."""
    ctx.testenv.setup(ctx)


@task(setup_testenv)
@with_stdout_header
def install_requirements(ctx):
    """Install all requirements, including tools used by Roberto."""
    if ctx.package_manager == "conda":
        install_requirements_conda(ctx)
    elif ctx.package_manager == "pip":
        install_requirements_pip(ctx)
    else:
        raise NotImplementedError


@task(install_requirements, sanitize_git, write_version)
@with_stdout_header
def lint_static(ctx):
    """Run static linters."""
    execute_tools(ctx, LintStatic)


@task(install_requirements, sanitize_git, write_version)
@with_stdout_header
def build_inplace(ctx):  # pylint: disable=too-many-branches
    """Build software in-place and update environment variables."""
    execute_tools(ctx, BuildInPlace)


@task(build_inplace)
@with_stdout_header
def test_inplace(ctx):
    """Run tests in-place."""
    execute_tools(ctx, TestInPlace)


@task(test_inplace)
@with_stdout_header
def upload_coverage(ctx):
    """Upload coverage reports, if explicitly enabled in config."""
    if ctx.upload_coverage:
        execute_tools(ctx, UploadCoverage)
    else:
        print("Coverage upload disabled by configuration.")


@task(build_inplace)
@with_stdout_header
def lint_dynamic(ctx):
    """Run dynamic linters."""
    execute_tools(ctx, LintDynamic)


@task(install_requirements, build_inplace, write_version)
@with_stdout_header
def build_docs(ctx):
    """Build documentation."""
    execute_tools(ctx, BuildDocs)


@task(install_requirements, build_docs)
@with_stdout_header
def upload_docs(ctx):
    """Upload documentation."""
    execute_tools(ctx, UploadDocs)


@task(install_requirements, write_version)
@with_stdout_header
def build_packages(ctx):
    """Build software package(s)."""
    execute_tools(ctx, BuildPackage)


@task(install_requirements, build_packages)
@with_stdout_header
def deploy(ctx):
    """Run all deployment tasks."""
    execute_tools(ctx, Deploy)


@task()
@with_stdout_header
def nuke_testenv(ctx):
    """Purge the test environment and stale source files. USE AT YOUR OWN RISK."""
    ctx.testenv.nuke(ctx)


@task(lint_static, test_inplace, upload_coverage, lint_dynamic, build_docs, default=True)
def quality(ctx):  # pylint: disable=unused-argument
    """Run all quality assurance tasks: linting and in-place testing."""
    print("\033[0;92m All QA checks passed!\033[0;0m")


@task(quality, upload_docs, deploy)
def robot(ctx):  # pylint: disable=unused-argument
    """Run all tasks, except nuke_virtual_env."""
    print("\033[0;92m All tasks completed!\033[0;0m")



"""Autogenerated version information file.

Manual changes to this file will be overwritten.

The output of git describe was: ``2.0.3``

The derived version is: ``2.0.3``
"""
__version__ = '2.0.3'
DEV_CLASSIFIER = 'Development Status :: 5 - Production/Stable'



# Collection of configurable development workflows
# Copyright (C) 2011-2019 The Roberto Development Team
#
# This file is part of Roberto.
#
# Roberto is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Roberto is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --
"""Define the Roberto's configuration."""

import os
import subprocess
try:
    from importlib_resources import open_text
except ImportError:
    from importlib.resources import open_text

from invoke.config import Config, merge_dicts, DataProxy
import yaml

from .utils import parse_git_describe
from .tools import initialize_tools
from .testenvs import init_testenv


class RobertoConfig(Config):
    """A specialized Invoke configuration for Roberto.

    The main modifications in behavior w.r.t. the vanilla Config are:

    - The `roberto` prefix.
    - A project config file `.roberto.*` can be loaded from the current directory.
    - Default configuration, which can be extended overridden by user.
    - Config finalization, filling in some blanks with sensible defaults.

    """

    prefix = 'roberto'

    def load_base_conf_files(self):
        """Load also the local project config file."""
        Config.load_base_conf_files(self)
        # Always try to load the project config file.
        self._load_file(prefix="project", merge=False)

    def set_project_location(self, path: str):
        """Override the default mechanism of Invoke to find project config.

        Parameters
        ----------
        path:
            This argument is ignored and the local directory is used instead.

        """
        # Impose our own project location, config file prefixed with dot.
        self._set(_project_prefix=os.path.join(os.getcwd(), '.'))
        self._set(_project_path=None)
        self._set(_project_found=None)
        self._set(_project={})

    def load_shell_env(self):
        """Call _finalize after Invoke has loaded the complete config."""
        Config.load_shell_env(self)
        self._finalize()

    def _finalize(self):
        """Derive some config variables and fill in defaults for convenience."""
        # Check if essential configuration is present.
        if self.project.name is None:
            raise TypeError("No project name defined in the configuration. Missing .roberto.yaml?")

        # Prepare download directory
        self.download_dir = os.path.expandvars(os.path.expanduser(self.download_dir))
        if not os.path.isdir(self.download_dir):
            os.makedirs(self.download_dir)

        # Package default options
        self.project.packages = [
            DataProxy.from_data(package) for package in self.project.packages]
        for package in self.project.packages:
            if 'path' not in package:
                package.path = '.'
            package.abspath = os.path.abspath(package.path)
            if 'name' not in package:
                package.name = self.project.name
            if 'tools' not in package:
                package.tools = []
            if 'conda_name' in package:
                raise RuntimeError("conda_name is no longer supported. Use dist_name instead.")
            if 'dist_name' not in package:
                package.dist_name = package.name
            # Check if all tools exist
            for toolname in package.tools:
                if toolname not in self.tools:
                    raise ValueError(f"Unknown Roberto tool: {toolname}")

        # Replace the testenv string by an object
        init_testenv(self)

        # Initialize tools
        initialize_tools(self.tools)

    @staticmethod
    def global_defaults() -> dict:
        """Set the global default configuration, before loading any other config."""
        defaults = Config.global_defaults()

        # Load default configuration
        with open_text('roberto', 'default_config.yaml') as f:
            defaults = merge_dicts(defaults, yaml.safe_load(f))

        # Git version and branch information
        try:
            git_describe = subprocess.run(
                ['git', 'describe', '--tags'],
                stdout=subprocess.PIPE, stderr=subprocess.DEVNULL,
                check=True).stdout.decode('utf-8')
        except subprocess.CalledProcessError:
            # May fail, e.g. when there are no tags.
            git_describe = '0.0.0-0-notag'
        defaults['git'].update(parse_git_describe(git_describe))

        # First try to get a decent branch name
        branch = subprocess.run(
            ["git", "rev-parse", "--abbrev-ref", "HEAD"],
            stdout=subprocess.PIPE, stderr=subprocess.DEVNULL,
            check=True).stdout.decode('utf-8').strip()
        # If that failed, try to get the tag
        if branch == 'HEAD':
            try:
                branch = subprocess.run(
                    ["git", "describe", "--tags", "--exact-match"],
                    stdout=subprocess.PIPE, stderr=subprocess.DEVNULL,
                    check=True).stdout.decode('utf-8').strip()
            except subprocess.CalledProcessError:
                # Final attempt, just the sha.
                try:
                    branch = subprocess.run(
                        ["git", "rev-parse", "HEAD"],
                        stdout=subprocess.PIPE, stderr=subprocess.DEVNULL,
                        check=True).stdout.decode('utf-8').strip()
                except subprocess.CalledProcessError:
                    branch = '__nogit__'
        defaults['git']['branch'] = branch

        return defaults



#!/usr/bin/env python3
# Collection of configurable development workflows
# Copyright (C) 2011-2019 The Roberto Development Team
#
# This file is part of Roberto.
#
# Roberto is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Roberto is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --
"""roberto setup script.

If you are not familiar with setup.py, just use pip instead:

    pip install roberto --user --upgrade

Alternatively, you can install from source with

    ./setup.py install --user
"""

import os

from setuptools import setup


NAME = 'roberto'


def get_version_info():
    """Read __version__ and DEV_CLASSIFIER from version.py, using exec, not import."""
    with open(os.path.join(NAME, 'version.py'), 'r') as f:
        myglobals = {"__name__": f"{NAME}.version"}
        exec(f.read(), myglobals)  # pylint: disable=exec-used
    return myglobals['__version__'], myglobals['DEV_CLASSIFIER']


def load_readme():
    """Load README for display on PyPI."""
    with open('README.rst') as f:
        return f.read()


VERSION, DEV_CLASSIFIER = get_version_info()


setup(
    name=NAME,
    version=VERSION,
    package_dir={NAME: NAME},
    packages=[NAME, NAME + '.test'],
    description='Collection of configurable development workflows',
    long_description=load_readme(),
    author='HORTON-ChemTools Dev Team',
    author_email='horton.chemtools@gmail.com',
    url='https://github.com/theochem/roberto',
    include_package_data=True,
    install_requires=[
        'invoke', 'pyyaml', 'importlib_resources; python_version < "3.7"'],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': ['rob = roberto.__main__:main']
    },
    classifiers=[
        DEV_CLASSIFIER,
        'Environment :: Console',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
)



# Collection of configurable development workflows
# Copyright (C) 2011-2019 The Roberto Development Team
#
# This file is part of Roberto.
#
# Roberto is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# Roberto is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
# --
"""Unit tests roberto.requirements."""

import os

from ..requirements import compute_req_hash


def test_req_hash(tmpdir):
    tmpdir = str(tmpdir)  # for python-3.5 compatibility
    req_items = set(["conda-build", "anaconda-client", "conda-verify"])
    # Use a fake but safe recipe dir
    fn_foo = os.path.join(tmpdir, "foo")
    with open(fn_foo, "w") as f:
        f.write("bar")
    req_fns = set([fn_foo])
    hash1 = compute_req_hash(req_items, req_fns)
    assert len(hash1) == 64
    req_items.add("1")
    hash2 = compute_req_hash(req_items, req_fns)
    assert len(hash2) == 64
    assert hash1 != hash2
    req_items.add("aaa")
    hash3 = compute_req_hash(req_items, req_fns)
    assert len(hash3) == 64
    assert hash1 != hash3
    assert hash2 != hash3
    fn_egg = os.path.join(tmpdir, "egg")
    with open(fn_egg, "w") as f:
        f.write("spam")
    req_fns.add(fn_egg)
    hash4 = compute_req_hash(req_items, req_fns)
    assert len(hash4) == 64
    assert hash1 != hash4
    assert hash2 != hash4
    assert hash3 != hash4
